from typing import Any
import inspect

from django.db import models
from django.apps import apps

from ledgers.models.ledgers import CumulativeLedgerMovementRecord, CumulativeLedgerTotalsRecord
from ledgers.ledger.fields.dimensions import Dimension
from ledgers.ledger.fields.resources import Resource


class LedgerMeta:
    abstract = False
    movements_class = CumulativeLedgerMovementRecord
    totals_class = CumulativeLedgerTotalsRecord


class CumulativeLedgerMeta(type):
    """ Cumulative Ledger Metaclass.

    This metaclass constructs a CumulativeLedger type from declared fields.
    Fields can be declared as:
        - dimensions
        - resources
        - attributes

    Dimensions — fields which, when combined, represent a unique slice to store movement records.
    Resources — fields which will be aggregated and totalled across many movement records.
    Attributes — fields which may store additional data bound to movement record.

    From declared fields, two Django Models are designed:
    - Movements table (inherits from ledgers.models.ledger.cumulative.CumulativeLedgerMovementRecord)
    - Totals table (inherits from ledgers.models.ledger.cumulative.CumulativeLedgerTotalsRecord)

    e.g. a cumulative ledger which stores item balances across warehouses.

    Dimensions:
        - warehouse (ref to Warehouse table)
        - item      (ref to Item table)
    Resources:
        - amount    (countable data type (int/float/numeric))
    Attributes:
        - comment   (arbitraty data)

    Movements table
        - columns:
            - period                    <-|
            - registrator_content_type    |
            - registrator_id              |- inherited from CumulativeLedgerMovementRecord (1)
            - line_index                  |
            - enabled                     |
            - action                    <-|
            - warehouse_id
            - item_id
            - amount
            - comment
        - indexes:
            - period                                                               <-|
            - registrator_content_type, registrator_id, line_index, enabled          |
            - registrator_id                                                         |- inh. as (1)
            - period, registrator_content_type, registrator_id, action, enabled      |
            - period, action, enabled                                              <-|
            - period, action, enabled, warehouse_id, item_id
        - constraints:
            - period, registrator_content_type, registrator_id, line_index   <- inh. as (1)

    Totals table
        - columns:
            - period        <- inherited from CumulativeLedgerTotalsRecord (2)
            - warehouse_id
            - item_id
            - total_amount
        - indexes:
            - period                <- inh. as (2)
            - warehouse_id, item_id
        - constraints:
            - period, warehouse_id, item_id

    Movements and Totals Models are registered at django ORM, and added to new class as protected attributes.

    # TODO: methods to access data
    """

    @classmethod
    def movements_model_factory(cls,
                                movements_model_name: str,
                                movements_model_bases: tuple[type],
                                is_abstract: bool,
                                dimensions: dict[str, Dimension],
                                resources: dict[str, Resource],
                                attributes: dict[str, models.Field],
                                extra_python_attributes: dict[str, Any]):

        meta = type(
            '_meta',
            (CumulativeLedgerMovementRecord.Meta,),
            {
                'abstract': is_abstract,
                'indexes': [
                    *CumulativeLedgerMovementRecord.Meta.indexes,
                    models.Index(
                        fields=[
                            'period', 'enabled', *dimensions.keys()
                        ]
                    ),
                ],
            }
        )

        model_fields = {}

        for name, dimension in dimensions.items():
            model_fields[name] = dimension.django_field

        for name, resource in resources.items():
            model_fields[name] = resource.django_field

        model_fields.update(attributes)

        if not len(model_fields) == sum(map(len, [dimensions, resources, attributes])):
            # TODO: exception when same named fields
            raise ValueError

        model_attrs = {
            'Meta': meta,
            **model_fields,
            **extra_python_attributes,
        }

        model = type(
            movements_model_name,
            movements_model_bases,
            model_attrs,
        )

        return model

    @classmethod
    def totals_model_factory(cls,
                             totals_model_name: str,
                             totals_model_bases: tuple[type],
                             is_abstract: bool,
                             dimensions: dict[str, Dimension],
                             resources: dict[str, Resource],
                             extra_python_attributes: dict[str, Any]):
        meta = type(
            '_meta',
            (CumulativeLedgerTotalsRecord.Meta,),
            {
                'abstract': is_abstract,
                'indexes': [
                    *CumulativeLedgerTotalsRecord.Meta.indexes,
                    models.Index(
                        fields=[
                            'period', *dimensions.keys(),
                        ]
                    ),
                ],
                'constraints': [
                    *(CumulativeLedgerTotalsRecord.Meta.constraints
                      if hasattr(CumulativeLedgerTotalsRecord.Meta, 'constraints') else []),
                    models.UniqueConstraint(
                        'period', *dimensions.keys(),
                        name=f'unique_dimensions_{totals_model_name}'
                    ),
                ]
            }
        )

        model_fields = {}

        for name, dimension in dimensions.items():
            model_fields[name] = dimension.django_field

        for name, resource in resources.items():
            model_fields[f'{name}_total'] = resource.django_field

        if not len(model_fields) == sum(map(len, [dimensions, resources])):
            # TODO: exception when same named fields
            raise ValueError

        model_attrs = {
            'Meta': meta,
            **model_fields,
            **extra_python_attributes,
        }

        model = type(
            totals_model_name,
            totals_model_bases,
            model_attrs,
        )

        return model

    @staticmethod
    def extract_contributable(attrs):
        def _has_contribute_to_class(value):
            return not inspect.isclass(value) and hasattr(value, "contribute_to_class")

        dimensions = {}
        resources = {}
        attributes = {}
        non_contributables = {}

        for obj_name, obj in attrs.items():
            if isinstance(obj, Dimension):
                dimensions[obj_name] = obj
                obj.set_name(obj_name)
            elif isinstance(obj, Resource):
                resources[obj_name] = obj
                obj.set_name(obj_name)
            elif _has_contribute_to_class(obj):
                attributes[obj_name] = obj
            else:
                non_contributables[obj_name] = obj

        return dimensions, resources, attributes, non_contributables

    def __new__(cls, name, bases, attrs, **kwargs):
        # Copy python metadata to new attrs
        module = attrs.pop("__module__")
        new_attrs = {"__module__": module}

        classcell = attrs.pop("__classcell__", None)
        if classcell is not None:
            new_attrs["__classcell__"] = classcell

        # Save meta and contributable attributes
        attr_meta = attrs.pop("LedgerMeta", LedgerMeta)
        abstract = attr_meta.abstract
        # Extract contributable
        dimensions, resources, attributes, non_contributables = cls.extract_contributable(attrs)

        # TODO: validate dim, res, attr after parse

        # Copy non-contributables
        new_attrs.update(non_contributables)

        # Create new class
        new_class = super().__new__(cls, name, bases, new_attrs, **kwargs)

        # Create models
        mvm_model = new_class.movements_model_factory(f'{name}_movements', (attr_meta.movements_class,), abstract,
                                                      dimensions, resources, attributes, {"__module__": module})
        tot_model = new_class.totals_model_factory(f'{name}_totals', (attr_meta.totals_class,), abstract, dimensions,
                                                   resources, {"__module__": module})

        # Add models to class
        setattr(new_class, 'mvm_model', mvm_model)
        setattr(new_class, 'tot_model', tot_model)

        # Add dimensions to class
        setattr(new_class, '_dimensions', dimensions.values())
        setattr(new_class, '_resources', resources.values())

        if not abstract:
            ledgers_app_config = apps.get_app_config('ledgers')

            app_config = apps.get_containing_app_config(module)
            app_label = app_config.label

            ledgers_app_config.register_ledger(app_label, new_class)

        return new_class



from django.db import models
from django.utils.translation import gettext_lazy as _


class PeriodicLedgerRecord(models.Model):
    """Periodic ledger record abstract class.

    These records have attributes:
    - period — a period in time when record becomes "active"
    """
    class Meta:
        abstract = True

        # recommended indexes (add to subclasses)
        indexes = [
            models.Index(
                fields=["period", ],
            ),
        ]

    period = models.DateTimeField(
        null=False,
        blank=False,
        verbose_name=_('record\'s period')
    )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ledgers.models import Document

from django.db import models, transaction
from django.utils.translation import gettext_lazy as _
from django.contrib.contenttypes.models import ContentType

from ledgers.utils.decorators import concurrent_transaction


class DocumentIteratorManager(models.Manager):

    @concurrent_transaction
    def _next_iterator(self, content_type_id: str, prefix: str, increment: int = 1) -> int:
        """
        To save modifications "outside" possible atomic block — this method is called in separate thread.
        This might result in a deadlock if "parent" thread locks this table.

        :param content_type_id: contenttype id
        :type content_type_id: str
        :param prefix: prefix lookup
        :type prefix: str
        :param increment: increment amount
        :type increment: int

        :return: next iterator from database
        :rtype: int
        """

        iterator, created = self.get_queryset().select_for_update().get_or_create(
            model_id=content_type_id,
            prefix=prefix,
            defaults={
                'number': increment,
            }
        )

        if created:
            return 1

        result = iterator.number

        iterator.number = models.F('number') + increment
        iterator.save()

        return result

    def next_iterator(self, document_model: 'Document', prefix: str, increment: int = 1) -> int:
        """
        This method gets next iterator for given document model and prefix.

        To solve race conditions, filtered row is locked, as there is no native Django
        '... RETURNING ...' implementation.

        :param document_model: document model
        :type document_model: ledgers.models.document.doc.Document
        :param prefix: prefix lookup
        :type prefix: str
        :param increment: increment amount
        :type increment: int

        :return: next iterator from database
        :rtype: int
        """

        content_type = ContentType.objects.get_for_model(document_model)

        return self._next_iterator(content_type.pk, prefix, increment)

    @transaction.atomic
    def rebuid(self):
        self.get_queryset().select_for_update().delete()

        iterators = []

        for descriptor, document_model in self.model.app_config.documents.items():
            content_type_id = ContentType.objects.get_for_model(document_model).pk

            qs = document_model.objects.values('number_prefix').annotate(
                max_iterator=models.Max('number_iterator')
            )

            for item in qs:
                iterators.append(
                    self.model(
                        model_id=content_type_id,
                        prefix=item['prefix'],
                        number=item['max_iterator'],
                    )
                )

        self.get_queryset().bulk_create(iterators)


class DocumentIterator(models.Model):
    """
    This model stores last number iterator for given document model and prefix.
    """

    class Meta:
        indexes = [
            models.Index(fields=['model', 'prefix']),
        ]
        constraints = [
            models.UniqueConstraint(
                fields=['model', 'prefix'],
                name='u_model_prefix',
            ),
        ]

    model = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        related_name='iterators',
        null=False,
        blank=False,
        editable=False,
        verbose_name=_('document model'),
    )
    prefix = models.CharField(
        max_length=3,
        null=False,
        blank=False,
        editable=False,
        verbose_name=_('document prefix'),
    )
    number = models.PositiveIntegerField(
        null=False,
        blank=False,
        editable=False,
        verbose_name=_('document number (as integer)'),
    )

    objects = DocumentIteratorManager()



from django.db import models

from .base import LedgerField


class Dimension(LedgerField):
    pass


class ForeignKey(Dimension):

    def __init__(self, to, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._to = to

    @property
    def django_field(self):
        return models.ForeignKey(
            to=self._to,
            on_delete=models.SET_NULL if self._null else models.CASCADE,
            null=self._null,
            blank=self._null,
            related_name=f'%(app_label)s_%(class)s',
        )

    @property
    def db_name(self) -> str:
        return f'{self._name}_id'



from django.db import models
from django.utils.translation import gettext_lazy as _
from django_cte import CTEManager

from ledgers.models.ledgers.bases import PeriodicLedgerRecord, DependentLedgerRecord
from ledgers.enum import CumulativeLedgerRecordActionEnum


class CumulativeLedgerMovementRecord(PeriodicLedgerRecord, DependentLedgerRecord):
    """
    Cumulative ledger movement record abstract class.

    These records have attributes:
    - action — specifies whether this record adds or substracts cumulatable value
    """

    class Meta:
        abstract = True

        # recommended indexes (add to subclasses)
        indexes = [
            *PeriodicLedgerRecord.Meta.indexes,
            *DependentLedgerRecord.Meta.indexes,
            models.Index(
                fields=["period", "registrator_content_type", "registrator_id", "action", "enabled", ],
            ),
            models.Index(
                fields=["period", "action", "enabled", ],
            ),
            models.Index(
                fields=["period", "enabled", ],
            )
        ]

        # recommended constraints (copy to subclasses)
        constraints = [
            models.UniqueConstraint(
                fields=["period", "registrator_content_type", "registrator_id", "line_index", ],
                name="unique_record",
            )
        ]

    action = models.CharField(
        null=False,
        blank=False,
        max_length=1,
        choices=CumulativeLedgerRecordActionEnum.choices,
        editable=True,
        verbose_name=_('action'),
    )

    objects = CTEManager()



from django.db import models

from .base import LedgerField


class Resource(LedgerField):
    pass


class Integer(Resource):

    @property
    def django_field(self):
        return models.IntegerField(
            null=self._null,
            blank=self._null,
        )



from abc import ABCMeta

from django.db import models


class LedgerField(metaclass=ABCMeta):
    def __init__(self, null=True):
        self._null = null
        self._name = None

    @property
    def name(self) -> None | str:
        return self._name

    @property
    def django_field(self) -> models.Field:
        raise NotImplemented

    def set_name(self, value: str) -> None:
        self._name = value

    @property
    def db_name(self) -> str:
        return self._name



import datetime
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ledgers.models import Document

from django.utils import timezone
from datetime import UTC
from django_cte import With
from django.db import models, transaction
from django.utils.functional import classproperty
from django.contrib.contenttypes.models import ContentType
from django.db.models import functions
from django.db.models.sql.constants import LOUTER

from .meta import LedgerMeta, CumulativeLedgerMeta
from ledgers.models import CumulativeLedgerMovementRecord, CumulativeLedgerTotalsRecord
from ledgers.ledger.fields.dimensions import Dimension
from ledgers.ledger.fields.resources import Resource
from ledgers.enum import DocumentStateEnum, CumulativeLedgerRecordActionEnum


class CumulativeLedger(metaclass=CumulativeLedgerMeta):
    mvm_model: CumulativeLedgerMovementRecord
    tot_model: CumulativeLedgerTotalsRecord

    _dimensions: list[Dimension]
    _resources: list[Resource]
    _TOTALS_FINAL_DATE = timezone.make_aware(datetime.datetime.max.replace(microsecond=0), timezone=UTC)

    class LedgerMeta(LedgerMeta):
        abstract = True

    # noinspection PyMethodParameters
    @classproperty
    def dimensions(cls) -> list[Dimension]:
        return cls._dimensions

    # noinspection PyMethodParameters
    @classproperty
    def resources(cls) -> list[Resource]:
        return cls._resources

    # noinspection PyMethodParameters
    @classproperty
    def dimensions_db_names(cls) -> list[str]:
        return [dim.db_name for dim in cls.dimensions]

    # noinspection PyMethodParameters
    @classproperty
    def resources_db_names(cls) -> list[str]:
        return [res.db_name for res in cls.resources]

    @classmethod
    @transaction.atomic
    def _delete_previous_movements(cls, document: 'Document') -> set[tuple[Any]]:
        """
        This method deletes movements from ledger for specific document.

        :param document: document
        :type document: ledgers.models.documents.document.Document

        :return: affected keys for totals to be recalculated
        :rtype: set[tuple[Any]]
        """
        ct = ContentType.objects.get_for_model(type(document))

        qs = cls.mvm_model.objects.filter(registrator_content_type_id=ct.pk, registrator_id=document.pk)
        affected_keys_values = list(qs.order_by().values_list(*cls.dimensions_db_names).distinct())

        affected_keys = set()
        for key_dimensions in affected_keys_values:
            key_dimensions = tuple(key_dimensions)
            affected_keys.add(key_dimensions)

        qs.delete()

        return affected_keys

    @classmethod
    @transaction.atomic
    def _write_movements(cls, document: 'Document') -> set[tuple[Any]]:
        """
        This method writes movements in ledger for specific document.

        :param document: document
        :type document: ledgers.models.documents.document.Document

        :return: affected keys for totals to be recalculated
        :rtype: set[tuple[Any]]
        """
        affected_keys = set()

        if document.deleted or document.state == DocumentStateEnum.DRAFT:
            return affected_keys

        ct = ContentType.objects.get_for_model(type(document))

        # records for this document type must be implemented in subclass directly
        movements_records = getattr(cls, f'movements_for_{ct.model}')(document)
        cls.mvm_model.objects.bulk_create(movements_records)

        for movements_record in movements_records:
            key_dimensions = []

            for dim_db_name in cls.dimensions_db_names:
                value = getattr(movements_record, dim_db_name)
                key_dimensions.append(value)

            key_dimensions = tuple(key_dimensions)

            affected_keys.add(key_dimensions)

        return affected_keys

    @classmethod
    @transaction.atomic
    def calculate_totals(cls, keys: set[tuple[Any]]):
        """
        This method calculates and wtites totals for provided keys.

        :param keys: dimension's keys to recalculate totals for
        :type keys: set[tuple[Any]]
        """

        new_totals = []

        dimensions = [
            dict(
                zip(
                    cls.dimensions_db_names,
                    dimension_values
                )
            )
            for dimension_values in keys
        ]

        q = models.Q()
        for dimension in dimensions:
            q |= models.Q(**dimension)

        _ = cls.tot_model.objects.filter(q).select_for_update()

        remains = list(
            cls.movements_remains_qs(
                period_q=None,
                dimensions_q=q,
            ).values(
                *cls.dimensions_db_names,
                *[f'{res_db_name}_mvm_sum' for res_db_name in cls.resources_db_names],
            )
        )

        totals = {
            tuple(row[dim_db_name] for dim_db_name in cls.dimensions_db_names): {
                f'{res_db_name}_total': row[f'{res_db_name}_mvm_sum']
                for res_db_name
                in cls.resources_db_names
            }
            for row
            in remains
        }

        for dimension in dimensions:
            # noinspection PyCallingNonCallable
            new_totals.append(
                cls.tot_model(
                    period=cls._TOTALS_FINAL_DATE,
                    **dimension,
                    **totals[tuple(dimension.values())],
                )
            )

        cls.tot_model.objects.bulk_create(
            new_totals,
            update_conflicts=True,
            update_fields=[f'{res_db_name}_total' for res_db_name in cls.resources_db_names],
            unique_fields=['period', *cls.dimensions_db_names],
        )

    @classmethod
    def movements_remains_qs(cls, period_q: models.Q = None, dimensions_q: models.Q = None,
                             over_dimensions: list[str] = None, resourses: list[str] = None):

        period_q = period_q or models.Q()
        dimensions_q = dimensions_q or models.Q()
        over_dimensions = over_dimensions or cls.dimensions_db_names
        resourses = resourses or cls.resources_db_names

        qs = cls.mvm_model.objects.filter(
            period_q,
            dimensions_q,
            enabled=True,
        ).values(*over_dimensions).annotate(
            **{
                f'{res}_mvm_sum': (
                        functions.Coalesce(
                            models.Sum(
                                res,
                                filter=models.Q(action=CumulativeLedgerRecordActionEnum.INCOME)
                            ),
                            0
                        ) -
                        functions.Coalesce(
                            models.Sum(
                                res,
                                filter=models.Q(action=CumulativeLedgerRecordActionEnum.OUTCOME)
                            ),
                            0
                        )
                )
                for res
                in resourses
            },
        )
        return qs

    @classmethod
    def remains_qs(cls, period: datetime.datetime = None, dimensions_q: models.Q = None, resourses: list[str] = None):
        """
        Algorithm to query remains for arbitraty date `date`:
        - get value from "totals table"
        - get all records from "movements table" between `date` and datetime.datetime.max
        - aggregate(sum) records from movements table
        - add "movements table"'s aggregates from "totals table"'s value
        """

        if not period:
            period = timezone.now()

        if not dimensions_q:
            dimensions_q = models.Q()

        if not resourses:
            resourses = cls.resources_db_names

        dimensions = cls.dimensions_db_names

        movements_cte = With(
            cls.movements_remains_qs(
                period_q=models.Q(period__gte=period),
                dimensions_q=dimensions_q,
                over_dimensions=dimensions,
                resourses=resourses,
            )
        )

        qs = movements_cte.join(
            cls.tot_model.objects.filter(
                dimensions_q,
                period=cls._TOTALS_FINAL_DATE,
            ),
            **{dim: getattr(movements_cte.col, dim) for dim in dimensions},
            _join_type=LOUTER,
        ).with_cte(movements_cte).values(*dimensions).annotate(
            **{
                f'{res}_remains': models.F(f'{res}_total') - functions.Coalesce(
                    getattr(movements_cte.col, f'{res}_mvm_sum'),
                    0,
                )
                for res
                in resourses
            },
        )

        return qs

    @classmethod
    def _bulk_delete_previous_records(cls, documents: list['Document']) -> set[tuple[Any]]:
        """
        This method deletes movements from ledger for a list of documents.

        :param documents: documents
        :type documents: list[ledgers.models.documents.document.Document]

        :return: affected keys for totals to be recalculated
        :rtype: set[tuple[Any]]
        """
        ct = ContentType.objects.get_for_model(type(documents[0]))

        qs = cls.mvm_model.objects.filter(content_type=ct, object_id__in=[d.pk for d in documents])
        affected_keys_values = list(qs.order_by().values_list(*cls.dimensions_db_names).distinct())

        affected_keys = set()
        for key_dimensions in affected_keys_values:
            key_dimensions = tuple(key_dimensions)
            affected_keys.add(key_dimensions)

        qs.delete()

        return affected_keys

    @classmethod
    @transaction.atomic
    def _bulk_write_movements(cls, documents: list['Document']) -> set[tuple[Any]]:
        """
        This method writes movements in ledger for a list of documents.

        :param documents: documents
        :type documents: list[ledgers.models.documents.document.Document]

        :return: affected keys for totals to be recalculated
        :rtype: set[tuple[Any]]
        """
        affected_keys = set()

        ct = ContentType.objects.get_for_model(type(documents[0]))

        movements_records = []
        for document in documents:

            if document.deleted or document.state == DocumentStateEnum.DRAFT:
                continue

            # TODO: call bulk_movements_for_...(documents) if implemented by ledger
            movements_records.extend(getattr(cls, f'movements_for_{ct.model}')(document))

        cls.mvm_model.objects.bulk_create(movements_records)

        for movements_record in movements_records:
            key_dimensions = []

            for dim_db_name in cls.dimensions_db_names:
                value = getattr(movements_record, dim_db_name)
                key_dimensions.append(value)

            key_dimensions = tuple(key_dimensions)

            affected_keys.add(key_dimensions)

        return affected_keys

    @classmethod
    @transaction.atomic
    def write(cls, document):

        deleted_keys = cls._delete_previous_movements(document)
        created_keys = cls._write_movements(document)

        affected_keys = deleted_keys.union(created_keys)

        cls.calculate_totals(affected_keys)

    @classmethod
    @transaction.atomic
    def bulk_write(cls, documents):
        # TODO: assert documents are homogenous

        deleted_keys = cls._bulk_delete_previous_records(documents)
        created_keys = cls._bulk_write_movements(documents)

        affected_keys = deleted_keys.union(created_keys)

        cls.calculate_totals(affected_keys)



from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy


class DocumentStateEnum(models.TextChoices):
    DRAFT = 'D', _('Draft')
    ACTIVE = 'A', pgettext_lazy('document state', 'Active')


class CumulativeLedgerRecordActionEnum(models.TextChoices):
    INCOME = 'I', _('Income')
    OUTCOME = 'O', _('Outcome')


class CumulativeLedgerSettingsTotalsPeriodEnum(models.TextChoices):
    DAY = 'D', _('Day')
    WEEK = 'W', _('Week')
    MONTH = 'M', _('Month')
    QUARTER = 'Q', _('Quarter')
    YEAR = 'Y', _('Year')



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ledgers.models.documents import Document

from django.db import models, transaction
from django.utils.translation import gettext_lazy as _
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.postgres.indexes import HashIndex

from ledgers.utils.arrays import is_homogenous


class DependentLedgerManager(models.Manager):
    """ A manager subclass for dependent ledger.
    Manages write and bulk write in ledger for document(s).

    Adds methods:
    - write — called upon document save.
    - bulk_write — a write implementation for an aray of documents (with same content type!).
    """

    def get_registrator_content_type(self, document) -> ContentType:
        """ This method gets content type for document and checks whether it can write in this ledger.

        :param document: document which content type is requested.
        :type document: ledgers.models.document.doc.Document

        :raises:
            ValueError: if document cannot write in this ledger.

        :return: content type of document
        :rtype: django.contrib.contenttypes.models.ContentType
        """
        ct = ContentType.objects.get_for_model(type(document))

        descriptor = f'{ct.app_label}.{ct.model}'.lower()

        if descriptor not in self.model.ALLOWED_REGISTRATOR_DESCRIPTORS:
            raise ValueError(f'Document {descriptor} cannot write to ledger {self.model.__name__}!')
        else:
            return ct

    @transaction.atomic
    def write(self, document: 'Document') -> None:
        """ Creates and saves ledger records from document.
        MUST be called on document save IF this ledger had dependency with document.

        :param document: a Document to generate records.
        :type document: ledgers.models.document.doc.Document
        """

        # get content type of document
        ct = self.get_registrator_content_type(document=document)

        # delete previous records
        self.get_queryset().filter(registrator=document).delete()

        # call records creation method
        records = getattr(self, f'records_for_{ct.model}')(document)

        # save records to DB
        self.get_queryset().bulk_create(records)

    @transaction.atomic
    def bulk_write(self, documents: list['Document']):
        """ Bulk impormentation of `write` method for a list of Documents.

        :param documents: a list of Documents to generate records.
        :type documents: list[ledgers.models.document.doc.Document]
        """

        if not documents:
            raise ValueError('Documents were not provided.')

        # assert that `documents` is homogenous
        if not is_homogenous(documents):
            raise ValueError('Documents are not homogenous.')

        # get content type of documents
        ct = self.get_registrator_content_type(document=documents[0])

        # delete previous records of those documents
        self.get_queryset().filter(registrator__in=documents).delete()

        # if model has an optimized records creation method
        if hasattr(self, f'bulk_records_for_{ct.model}'):
            # call it
            records = getattr(self, f'bulk_records_for_{ct.model}')(documents)
        else:
            # else fallback to standart records creation method in cycle
            records = []

            write_fn = getattr(self, f'records_for_{ct.model}')

            for document in documents:
                records.extend(write_fn(document))

        # save records to DB
        self.get_queryset().bulk_create(records)


class DependentLedgerRecord(models.Model):
    """Dependent ledger record abstract class.

    These records have attributes:
    - registrator — a reference to an arbitrary subclass of abstracts.models.document.Document.
    Implemented via contenttypes Generic Foreign Key (registrator_content_type + registrator_id).
    - enabled — a flag to specify whether this record is active and can be used in aggregates.

    ALLOWED_REGISTRATOR_DESCRIPTORS — used to limit documents which may write in this ledger.
    """

    class Meta:
        abstract = True

        # recommended indexes (add to subclasses)
        indexes = [
            models.Index(
                fields=["registrator_content_type", "registrator_id", "line_index", "enabled", ],
            ),
            HashIndex(
                fields=["registrator_id", ],
            ),
        ]

        # recommended constraints (copy to subclasses)
        constraints = [
            models.UniqueConstraint(
                fields=["registrator_content_type", "registrator_id", "line_index", ],
                name="unique_record",
            )
        ]

    ALLOWED_REGISTRATOR_DESCRIPTORS: list[str] = None

    registrator_content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        null=False,
        blank=False,
        verbose_name=_('registrator content type'),
    )
    registrator_id = models.CharField(
        max_length=36,
        null=False,
        blank=False,
        verbose_name=_('registrator id'),
    )
    registrator = GenericForeignKey(
        'registrator_content_type',
        'registrator_id',
    )
    line_index = models.PositiveIntegerField(
        null=False,
        blank=False,
        verbose_name=_('line index across this document'),
    )
    enabled = models.BooleanField(
        null=False,
        blank=False,
        default=True,
        verbose_name=_('this record activity flag'),
    )

    objects = DependentLedgerManager()



import functools
from threading import Thread

from django.db import connection


class ThreadWithContext(Thread):
    def __init__(self, context=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.context = context or {}

    def run(self) -> None:
        res = None

        try:
            if self._target is not None:
                res = self._target(*self._args, **self._kwargs)
        except Exception as e:
            self.context['state'] = 'FAILED'
            self.context['exception'] = e
        else:
            self.context['state'] = 'SUCCESS'
            self.context['result'] = res
        finally:
            del self._target, self._args, self._kwargs


def concurrent_transaction(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        thread = ThreadWithContext(target=fn)
        thread.start()
        if thread.is_alive():
            thread.join()

        if thread.context['state'] == 'FAILED':
            raise thread.context['exception']
        else:
            return thread.context['result']

    # SQLite does not support concurrency
    if connection.vendor == 'sqlite':
        return fn
    else:
        return wrapper



from django.apps import AppConfig


class LedgersConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ledgers'

    def __init__(self, app_name, app_module):
        super(LedgersConfig, self).__init__(app_name, app_module)
        self.ledgers = {}
        self.documents = {}

    def register_ledger(self, app_label, ledger):
        ledger_name = ledger.__name__.lower()
        self.ledgers[f'{app_label}.{ledger_name}'] = ledger

    def register_document(self, document):
        doc_name = document.__name__.lower()
        self.documents[f'{document._meta.app_label}.{doc_name}'] = document

    def ready(self):
        from ledgers.models import Document

        for model in self.apps.get_models():
            if issubclass(model, Document):
                self.register_document(model)




# Generated by Django 5.0a1 on 2023-10-17 12:48

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentIterator',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('prefix', models.CharField(editable=False, max_length=3, verbose_name='document prefix')),
                ('number', models.PositiveIntegerField(editable=False, verbose_name='document number (as integer)')),
                ('model', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='iterators', to='contenttypes.contenttype', verbose_name='document model')),
            ],
            options={
                'indexes': [models.Index(fields=['model', 'prefix'], name='ledgers_doc_model_i_bf0f1f_idx')],
            },
        ),
        migrations.AddConstraint(
            model_name='documentiterator',
            constraint=models.UniqueConstraint(fields=('model', 'prefix'), name='u_model_prefix'),
        ),
    ]



from typing import Sequence


def is_homogenous(seq: Sequence) -> bool:
    """Checks if a sequence is homogenous.

    :param seq: Mapping

    :raises:
        ValueError: if seq is not valid for test.

    :return: True if homogenous else False
    :rtype: bool
    """

    try:
        first_type = type(seq[0])
    except (KeyError, TypeError) as e:
        raise ValueError(f'Sequence {seq} is not valid for homogeneity test. Exception raised: {e}')

    return all(type(item) == first_type for item in seq)



from django_cte import CTEManager

from ledgers.models.ledgers.bases import PeriodicLedgerRecord


class CumulativeLedgerTotalsRecord(PeriodicLedgerRecord):
    """ Cumulative ledger totals records.

    This is an implementation of 1C's totals tables.
    Source:
        - https://its.1c.ru/db/pubapplied/content/123/hdoc (writing)
        - https://its.1c.ru/db/pubapplied/content/130/hdoc (querying)

    A brief explanation:

    To optimize querying remains for specific date, totals are calculated and stored.
    Totals are calculated and stored for final date (datetime.datetime.max).

    Algorithm to query remains for arbitraty date `date`:
    - get value from "totals table"
    - get all records from "movements table" between `date` and datetime.datetime.max
    - aggregate(sum) records from movements table
    - substract "movements table"'s aggregates from "totals table"'s value

    e.g.

    totals table:
    9999-12-31 23:59:59 — 500                       (_TOTALS_FINAL)

    and movements table:
    ...
    2023-04-20 00:00:00 — +100                      (m1)
    2023-04-21 00:00:00 — -250                      (m2)
    2023-04-25 00:00:00 — -350                      (m3)
    2023-04-26 00:00:00 — +100                      (m4)
    ...

    Then querying for date "2023-04-23 00:00:00" will be as:
     500             - ( -350 + 100 ) = 750
    (__TOTALS_FINAL)     (m3)  (m4)

    Totals are recalculated either:
    - manually
    - on movements tavle write

    This ensures that querying old data is not affected by the size of ledger movements,
    but rather from frequency of records per totals period.
    """

    class Meta(PeriodicLedgerRecord.Meta):
        abstract = True
        indexes = [
            *PeriodicLedgerRecord.Meta.indexes,
        ]

    objects = CTEManager()



from typing import Iterable, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from ledgers.ledger import CumulativeLedger

from collections import defaultdict

from django.apps import apps
from django.db import models, transaction
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.contrib.postgres.indexes import HashIndex

from ledgers.enum import DocumentStateEnum
from ledgers.models.documents import DocumentIterator


class DocumentQueryset(models.QuerySet):
    """
    Document Queryset class.

    Adds methods:
    - bulk_actualize_document_number — recalculates document number in bulk
    - bulk_write_ledger_records
    - bulk_resave_related_documents

    and extends some...

    """

    model: 'Document'

    def bulk_actualize_document_number(self, documents: Iterable['Document']) -> None:
        """
        This method recalculates document number for a list of documents inplace (without saving to DB).

        :param documents: a list of documents
        :type documents: Iterable[ledgers.models.document.doc.Document]
        """

        prefixes_and_docs: dict[str, list['Document']] = defaultdict(list)

        # noinspection PyArgumentList
        for _doc in sorted(documents, lambda d: d.execution_date):
            prefixes_and_docs[_doc.prefix].append(_doc)

        for prefix, documents in prefixes_and_docs.items():
            model = documents[0].__class__

            next_iterator = DocumentIterator.objects.next_iterator(model, prefix, len(documents))

            for index, doc in enumerate(documents):
                doc.number_prefix = prefix
                doc.number_iterator = next_iterator + index

                doc.number = f'{doc.number_prefix}-{str(doc.number_iterator).zfill(13)}'

    @transaction.atomic
    def bulk_write_ledger_records(self, documents: Iterable['Document']) -> None:
        """
        This method executes records creation for every related ledger.

        :param documents: a list of documents
        :type documents: Iterable[ledgers.models.document.doc.Document]
        """

        for ledger in self.model.DocumentMeta.related_ledgers:
            if isinstance(ledger, str):
                ledger = apps.get_app_config('ledgers').ledgers[ledger.lower()]

            ledger.bulk_write(documents)

    @transaction.atomic
    def bulk_resave_related_documents(self, documents: Iterable['Document']) -> None:
        """
        This method resaves related documents.

        :param documents: a list of documents
        :type documents: Iterable[ledgers.models.document.doc.Document]
        """

        for lookup in self.model.DocumentMeta.related_documents_lookups:
            related_field = getattr(self.model, lookup, None).field

            docs = list(
                related_field.model.objects.filter(
                    **{
                        f'{related_field.name}__in': documents,
                    }
                ).prefetch_related(
                    related_field.name
                ).all()
            )

            for doc in docs:
                parent_doc: Document = getattr(doc, related_field.name)
                doc.state = parent_doc.state
                doc.execution_date = parent_doc.execution_date

            related_field.model.objects.bulk_save(docs)

    @transaction.atomic
    def _prepare_for_bulk_create(self, objs):
        """
        This extension ensures that document number is generated for new documents.
        """
        super(DocumentQueryset, self)._prepare_for_bulk_create(objs)

        self.bulk_actualize_document_number(objs)

    @transaction.atomic
    def _after_bulk_update(self, documents: Iterable['Document']) -> None:
        """
        This method ensures that all ledger records and related documents will be created/updated after bulk update.

        :param documents: a list of updated documents
        :type documents: Iterable[ledgers.models.document.doc.Document]
        """

        self.bulk_write_ledger_records(documents)
        self.bulk_resave_related_documents(documents)

    @transaction.atomic
    def bulk_create(self, objs, *args, **kwargs):
        """
        This extension ensures that all ledger records and related documents will be created/updated after bulk create.
        """

        objs = super(DocumentQueryset, self).bulk_create(objs, *args, **kwargs)

        self._after_bulk_update(objs)

        return objs

    @transaction.atomic
    def bulk_update(self, objs, fields, batch_size=None):
        """
        This extension ensures that all ledger records and related documents will be created/updated after bulk create.
        """

        objs = super(DocumentQueryset, self).bulk_update(objs, fields, batch_size)

        self._after_bulk_update(objs)

        return objs

    def update(self, **kwargs):
        """
        Extends update method to write records to ledgers.
        """
        result = super().update(**kwargs)

        self._after_bulk_update(self)

        return result

    @transaction.atomic
    def delete(self):
        """
        Extends delete method to write "empty" records to ledgers at first, and then proceed to deletion.
        """
        self.update(_deleted=True)

        return super().delete()


class Document(models.Model):
    """
    Document abstract class.

    Documents are data aggregations, which atomically modifies database upon save,
    through ledgers and related documents.

    Document-specific attributes:
    - state — current state of document — active or draft.
        It is a convention, that draft document will not have any ledger records.
    - execution_date — a "date" of document
    - number_prefix, number_iterator, number — document's unique number and its elements

    """

    class Meta:
        abstract = True
        indexes = [
            HashIndex(fields=['number']),
            models.Index(fields=['number_prefix', 'number_iterator']),
            models.Index(fields=['execution_date']),
        ]

    class DocumentMeta:
        """
        A Document metadata class which stores:
        - related_ledgers — a list of ledgers, records to which will be written upon save
        - related_documents_lookups — a list of reverse lookups to this model to find "related" documents
        - default_prefix — a string which contains default prefix
        """
        # TODO: assert is an abstract.src.CumulativeLedger descriptor / class
        related_ledgers: list[Union['CumulativeLedger', str]] = []

        # TODO: assert is a reverse lookup to Document
        # TODO: assert no circular dependencies
        related_documents_lookups: list[str] = []

        default_prefix: str = 'DOC'

    state = models.CharField(
        null=False,
        blank=False,
        max_length=1,
        choices=DocumentStateEnum.choices,
        default=DocumentStateEnum.DRAFT,
        editable=True,
        verbose_name=_('document state'),
    )
    _deleted = models.BooleanField(
        default=False,
        null=False,
        blank=False,
        editable=False,
        verbose_name=_('marked as deleted')
    )
    execution_date = models.DateTimeField(
        null=False,
        blank=False,
        default=timezone.now,
        editable=True,
        verbose_name=_('document execution date'),
    )
    number_prefix = models.CharField(
        max_length=3,
        null=False,
        blank=False,
        editable=False,
        verbose_name=_('document prefix'),
    )
    number_iterator = models.PositiveIntegerField(
        null=False,
        blank=False,
        editable=False,
        verbose_name=_('document number (as integer)'),
    )
    number = models.CharField(
        max_length=13,
        null=False,
        blank=False,
        editable=False,
        verbose_name=_('document number'),
    )

    objects = models.Manager.from_queryset(DocumentQueryset)()

    @property
    def prefix(self) -> str:
        """
        Default prefix may be overriden by instance data.

        :return: document's prefix
        :rtype: str
        """

        return self.DocumentMeta.default_prefix

    def actualize_document_number(self) -> None:
        """
        Method recalculates document number.

        Document number template is XXX-0000000000000
                              prefix ^        ^ iterator
        """
        self.number_prefix = self.prefix

        self.number_iterator = DocumentIterator.objects.next_iterator(self.__class__, self.number_prefix)

        self.number = f'{self.number_prefix}-{str(self.number_iterator).zfill(13)}'

    def direct_save(self) -> None:
        """
        A fallback method to save document without triggering ledger / related doc saves.
        """
        super(Document, self).save()

    @transaction.atomic
    def write_ledgers_records(self) -> None:
        """
        This method writes ledger records for document.
        """
        for ledger in self.DocumentMeta.related_ledgers:
            if isinstance(ledger, str):
                ledger = apps.get_app_config('ledgers').ledgers[ledger.lower()]

            ledger.write(self)

    @transaction.atomic
    def resave_related_documents(self) -> None:
        """
        This method resaves all related documents.

        It changes 'state', 'execution_date' and '_deleted' arguments to match parent document.

        """

        for lookup in self.DocumentMeta.related_documents_lookups:
            related_field = getattr(self.__class__, lookup, None).field
            docs = list(
                related_field.model.objects.filter(
                    **{
                        f'{related_field.name}': self,
                    }
                ).prefetch_related(
                    related_field.name
                ).all()
            )

            for doc in docs:
                doc.state = self.state
                doc.execution_date = self.execution_date

            related_field.model.objects.bulk_update(docs, ['state', 'execution_date', '_deleted'])

    @transaction.atomic
    def save(self, *args, **kwargs):
        """
        Extends save method to ensure that instance has document number, writes ledger records and
        resaves related documents after original save.
        """
        if not self.number:
            self.actualize_document_number()

        super(Document, self).save(*args, **kwargs)

        self.write_ledgers_records()
        self.resave_related_documents()

    @transaction.atomic
    def delete(self, *args, **kwargs):
        """
        Extends delete method to write "empty" records to ledgers at first, and then proceed to deletion.
        """
        self._deleted = True
        self.save(update_fields=['_deleted'])

        super(Document, self).delete(*args, **kwargs)

    @property
    def deleted(self):
        return self._deleted

    def __str__(self):
        return f'{self.number}'
